from __future__ import annotations

from .utils import StrEnum


class Flavour(StrEnum):
    # The latest Classic version is always aliased to "classic".
    # The logic here is that should Classic not be discontinued
    # it will continue to be updated in place so that new Classic versions
    # will inherit the "_classic_" folder.  This means we won't have to
    # migrate Classic profiles either automatically or by requiring user
    # intervention for new Classic releases.
    retail = 'retail'
    vanilla_classic = 'vanilla_classic'
    burning_crusade_classic = 'classic'


class Strategy(StrEnum):
    default = 'default'
    latest = 'latest'
    any_flavour = 'any_flavour'
    version = 'version'


class ChangelogFormat(StrEnum):
    html = 'html'
    markdown = 'markdown'
    bbcode = 'bbcode'
    raw = 'raw'



from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import Awaitable, Callable, Iterable, Iterator, Mapping, Sequence, Set
from datetime import datetime, timedelta
import enum
from functools import partial, wraps
from itertools import chain, repeat
import os
from pathlib import Path, PurePath
import posixpath
from shutil import move as _move
from tempfile import mkdtemp
from typing import Any, Generic, Hashable, TypeVar, overload

_T = TypeVar('_T')
_U = TypeVar('_U')
_H = TypeVar('_H', bound=Hashable)


class StrEnum(str, enum.Enum):
    pass


class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, contents: str) -> None:
        self.entries = {
            k: v
            for e in contents.splitlines()
            if e.startswith('##')
            for k, v in (map(str.strip, e.lstrip('#').partition(':')[::2]),)
            if k
        }

    def __getitem__(self, key: str | tuple[str, ...]) -> str | None:
        if isinstance(key, tuple):
            return next(filter(None, map(self.entries.get, key)), None)
        else:
            return self.entries.get(key)

    @classmethod
    def from_addon_path(cls, path: Path, suffix: str = '.toc') -> TocReader:
        return cls((path / (path.name + suffix)).read_text(encoding='utf-8-sig', errors='replace'))


class cached_property(Generic[_T, _U]):
    def __init__(self, f: Callable[[_T], _U]) -> None:
        self.f = f

    @overload
    def __get__(self, o: None, t: type[_T] | None = ...) -> cached_property[_T, _U]:
        ...

    @overload
    def __get__(self, o: _T, t: type[_T] | None = ...) -> _U:
        ...

    def __get__(self, o: _T | None, t: type[_T] | None = None) -> cached_property[_T, _U] | _U:
        if o is None:
            return self
        else:
            o.__dict__[self.f.__name__] = v = self.f(o)
            return v


def bucketise(iterable: Iterable[_U], key: Callable[[_U], _T]) -> defaultdict[_T, list[_U]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket: defaultdict[_T, list[_U]] = defaultdict(list)
    for value in iterable:
        bucket[key(value)].append(value)
    return bucket


def chain_dict(
    keys: Iterable[_T], default: _U, *overrides: Iterable[tuple[_T, _U]]
) -> dict[_T, _U]:
    "Construct a dictionary from a series of two-tuple iterables with overlapping keys."
    return dict(chain(zip(keys, repeat(default)), *overrides))


def uniq(it: Iterable[_H]) -> list[_H]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))


def merge_intersecting_sets(it: Iterable[frozenset[_T]]) -> Iterator[frozenset[_T]]:
    "Recursively merge intersecting sets in a collection."
    many_sets = list(it)
    while many_sets:
        this_set = many_sets.pop(0)
        while True:
            for idx, other_set in enumerate(many_sets):
                if not this_set.isdisjoint(other_set):
                    this_set |= many_sets.pop(idx)
                    break
            else:
                break
        yield this_set


@overload
async def gather(it: Iterable[Awaitable[_U]], wrapper: None = ...) -> list[_U]:
    ...


@overload
async def gather(
    it: Iterable[Awaitable[_U]],
    wrapper: Callable[[Awaitable[_U]], Awaitable[_T]] = ...,
) -> list[_T]:
    ...


async def gather(
    it: Iterable[Awaitable[object]], wrapper: Callable[..., Awaitable[object]] | None = None
) -> Sequence[object]:
    if wrapper is not None:
        it = map(wrapper, it)
    return await asyncio.gather(*it)


@overload
def run_in_thread(
    fn: type[list[object]],
) -> Callable[[Iterable[_U]], Awaitable[list[_U]]]:
    ...


@overload
def run_in_thread(fn: Callable[..., _U]) -> Callable[..., Awaitable[_U]]:
    ...


def run_in_thread(fn: Callable[..., object]) -> Callable[..., Awaitable[object]]:
    @wraps(fn)
    def wrapper(*args: object, **kwargs: object):
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(None, partial(fn, *args, **kwargs))

    return wrapper


def tabulate(rows: Sequence[Sequence[object]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value: object):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows: Sequence[Sequence[str]]):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    rows = [tuple(map(apply_max_col_width, r)) for r in rows]
    head, *tail = rows

    base_template = '  '.join(f'{{{{{{0}}{w}}}}}' for w in calc_resultant_col_widths(rows))
    row_template = base_template.format(':<')
    table = '\n'.join(
        (
            base_template.format(':^').format(*head),
            base_template.format(f'0:-<').format(''),
            *(row_template.format(*r) for r in tail),
        )
    )
    return table


def move(src: str | os.PathLike[str], dest: str | os.PathLike[str]) -> Any:
    return _move(
        os.fspath(src),  # See https://bugs.python.org/issue32689
        dest,
    )


def trash(paths: Sequence[PurePath], *, dest: PurePath, missing_ok: bool = False) -> None:
    if not paths:
        return

    parent_folder = mkdtemp(dir=dest, prefix=f'deleted-{paths[0].name}-')
    for path in paths:
        try:
            move(path, dest=parent_folder)
        except (FileNotFoundError if missing_ok else ()):
            pass


def shasum(*values: object) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(map(str, values)).encode()).hexdigest()[:32]


def is_not_stale(path: Path, ttl: Mapping[str, float]) -> bool:
    "Check if a file is older than ``ttl``."
    return path.exists() and (
        (datetime.now() - datetime.fromtimestamp(path.stat().st_mtime)) < timedelta(**ttl)
    )


def find_addon_zip_base_dirs(names: Sequence[str]) -> Iterator[str]:
    "Find top-level folders in a list of ZIP member paths."
    for name in names:
        if name.count(posixpath.sep) == 1:
            head, tail = posixpath.split(name)
            if tail.startswith(head) and tail[-4:].lower() == '.toc':
                yield head


def make_zip_member_filter(base_dirs: Set[str]) -> Callable[[str], bool]:
    "Filter out items which are not sub-paths of top-level folders in a ZIP."

    def is_subpath(name: str):
        head, sep, _ = name.partition(posixpath.sep)
        return head in base_dirs if sep else False

    return is_subpath


def file_uri_to_path(file_uri: str) -> str:
    "Convert a file URI to a path that works both on Windows and *nix."
    from urllib.parse import unquote

    unprefixed_path = unquote(file_uri[7:])  # len('file://')
    # A slash is prepended to the path even when there isn't one there
    # on Windows.  The ``PurePath`` instance will inherit from either
    # ``PurePosixPath`` or ``PureWindowsPath``; this will be a no-op on POSIX.
    if PurePath(unprefixed_path[1:]).drive:
        unprefixed_path = unprefixed_path[1:]
    return unprefixed_path


def normalise_names(replace_delim: str) -> Callable[[str], str]:
    import string

    trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

    def normalise(value: str):
        return replace_delim.join(value.casefold().translate(trans_table).split())

    return normalise



"""
Create foreign key constraint on ``pkg_version_log``.

Revision ID: 9b8df9661cdb
Revises: 764fa963cc71
Create Date: 2021-09-25 18:29:27.577648

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '9b8df9661cdb'
down_revision = '764fa963cc71'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg_version_log', schema=None) as batch_op:
        batch_op.create_foreign_key(
            'fk_pkg_version_log_pkg_source_and_id',
            'pkg',
            ['pkg_source', 'pkg_id'],
            ['source', 'id'],
        )


def downgrade():
    with op.batch_alter_table('pkg_version_log', schema=None) as batch_op:
        batch_op.drop_constraint('fk_pkg_version_log_pkg_source_and_id', type_='foreignkey')



"""
Rename the 'origin' column to 'source'.
"""
from functools import partial

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8f6ba74cfa82'
down_revision = '3b37d2faccd2'
branch_labels = None
depends_on = None

batch_alter_table = partial(
    op.batch_alter_table,
    naming_convention={'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s'},
    schema=None,
)


def upgrade():
    with batch_alter_table('pkg') as batch_op:
        batch_op.alter_column('origin', new_column_name='source')

    with batch_alter_table('pkg_dep') as batch_op:
        batch_op.alter_column('pkg_origin', new_column_name='pkg_source')
        batch_op.drop_constraint('uq_id_per_foreign_key_constr', type_='unique')
        batch_op.create_unique_constraint(
            'uq_id_per_foreign_key_constr', ['id', 'pkg_source', 'pkg_id']
        )
        batch_op.drop_constraint('fk_pkg_dep_pkg_origin_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_dep_pkg_source_pkg', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )

    with batch_alter_table('pkg_folder') as batch_op:
        batch_op.alter_column('pkg_origin', new_column_name='pkg_source')
        batch_op.drop_constraint('fk_pkg_folder_pkg_origin_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_folder_pkg_source_pkg', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )

    with batch_alter_table('pkg_options') as batch_op:
        batch_op.alter_column('pkg_origin', new_column_name='pkg_source')
        batch_op.drop_constraint('fk_pkg_options_pkg_origin_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_options_pkg_source_pkg', 'pkg', ['pkg_source', 'pkg_id'], ['source', 'id']
        )


def downgrade():
    with batch_alter_table('pkg') as batch_op:
        batch_op.alter_column('source', new_column_name='origin')

    with batch_alter_table('pkg_dep') as batch_op:
        batch_op.alter_column('pkg_source', new_column_name='pkg_origin')
        batch_op.drop_constraint('uq_id_per_foreign_key_constr', type_='unique')
        batch_op.create_unique_constraint(
            'uq_id_per_foreign_key_constr', ['id', 'pkg_origin', 'pkg_id']
        )
        batch_op.drop_constraint('fk_pkg_dep_pkg_source_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_dep_pkg_origin_pkg', 'pkg', ['pkg_origin', 'pkg_id'], ['origin', 'id']
        )

    with batch_alter_table('pkg_folder') as batch_op:
        batch_op.alter_column('pkg_source', new_column_name='pkg_origin')
        batch_op.drop_constraint('fk_pkg_folder_pkg_source_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_folder_pkg_origin_pkg', 'pkg', ['pkg_origin', 'pkg_id'], ['origin', 'id']
        )

    with batch_alter_table('pkg_options') as batch_op:
        batch_op.alter_column('pkg_source', new_column_name='pkg_origin')
        batch_op.drop_constraint('fk_pkg_options_pkg_source_pkg', type_='foreignkey')
        batch_op.create_foreign_key(
            'fk_pkg_options_pkg_origin_pkg', 'pkg', ['pkg_origin', 'pkg_id'], ['origin', 'id']
        )



from __future__ import annotations

import asyncio

import aiohttp

CLIENT_ID = 'aa178904bdf5143e93ec'


async def get_codes(web_client: aiohttp.ClientSession):
    async with web_client.post(
        'https://github.com/login/device/code',
        headers={'Accept': 'application/json'},
        json={'client_id': CLIENT_ID},
        raise_for_status=True,
    ) as response:
        return await response.json()


async def poll_for_access_token(
    web_client: aiohttp.ClientSession, device_code: str, polling_interval: int = 5
):
    while True:
        async with web_client.post(
            'https://github.com/login/oauth/access_token',
            headers={'Accept': 'application/json'},
            json={
                'client_id': CLIENT_ID,
                'device_code': device_code,
                'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
            },
            raise_for_status=True,
        ) as response:
            response_json = await response.json()
            if 'error' in response_json:
                if response_json['error'] == 'authorization_pending':
                    await asyncio.sleep(polling_interval)
                else:
                    raise ValueError('Authorization failed', response_json)
            else:
                return response_json['access_token']



"""
Add UNIQUE constraint that was missing from ``pkg_dep``.
"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3b37d2faccd2'
down_revision = '7204944522b1'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg_dep') as batch_op:
        batch_op.create_unique_constraint(
            'uq_id_per_foreign_key_constr', ['id', 'pkg_origin', 'pkg_id']
        )


def downgrade():
    with op.batch_alter_table('pkg_dep') as batch_op:
        batch_op.drop_constraint('uq_id_per_foreign_key_constr', type_='unique')



from __future__ import annotations

from collections.abc import Callable as C
from functools import partial
from itertools import product
import json
from typing import Any

from aiohttp import ClientWebSocketResponse, WSServerHandshakeError
from aiohttp.test_utils import TestClient, TestServer
import pytest
from yarl import URL

from instawow.config import Config

try:
    from instawow_gui import json_rpc_server
except ImportError:
    pytestmark = pytest.mark.skip(reason='instawow_gui is not available')


dumps = partial(json.dumps, default=str)


@pytest.fixture
async def ws_client(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_global_config_values['config_dir']))
    app = await json_rpc_server.create_app()
    server = TestServer(app)
    async with TestClient(server) as client:
        yield client


@pytest.fixture
async def ws(ws_client: TestClient):
    async with ws_client.ws_connect('/api', origin=str(ws_client.make_url(''))) as ws:
        yield ws


@pytest.mark.asyncio
async def test_no_origin_api_request_rejected(ws_client: TestClient):
    with pytest.raises(WSServerHandshakeError):
        async with ws_client.ws_connect('/api'):
            pass


@pytest.mark.parametrize(
    'transform',
    [
        lambda u: u.with_scheme('ftp'),
        lambda u: u.with_host('example.com'),
        lambda u: u.with_port(21),
    ],
)
@pytest.mark.asyncio
async def test_disparate_origin_api_request_rejected(
    ws_client: TestClient,
    transform: C[[URL], URL],
):
    with pytest.raises(WSServerHandshakeError):
        async with ws_client.ws_connect('/api', origin=str(transform(ws_client.make_url('')))):
            pass


@pytest.mark.asyncio
async def test_write_config(
    request: pytest.FixtureRequest,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
    ws: ClientWebSocketResponse,
):
    config_values = {**iw_config_values, 'profile': request.node.name}
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'config/write',
        'params': {**config_values, 'infer_game_flavour': False},
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['id'] == request.node.name
    assert Config(**rpc_response['result']) == Config(
        global_config=iw_global_config_values, **config_values
    )


@pytest.mark.asyncio
async def test_write_config_with_invalid_params(
    request: pytest.FixtureRequest,
    iw_config_values: dict[str, Any],
    ws: ClientWebSocketResponse,
):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'config/write',
        'params': {
            **iw_config_values,
            'game_flavour': 'strawberry',
            'infer_game_flavour': False,
        },
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['id'] == request.node.name
    assert rpc_response['error']
    assert rpc_response['error']['code'] == -32602
    assert rpc_response['error']['message'] == 'Invalid method parameter(s).'
    assert rpc_response['error']['data'] == [
        {
            'loc': ['game_flavour'],
            'msg': "value is not a valid enumeration member; permitted: 'retail', 'vanilla_classic', 'classic'",
            'type': 'type_error.enum',
            'ctx': {'enum_values': ['retail', 'vanilla_classic', 'classic']},
        }
    ]


@pytest.mark.asyncio
async def test_install_with_invalid_params(
    request: pytest.FixtureRequest,
    ws: ClientWebSocketResponse,
):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'install',
        'params': {},
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['error'] and rpc_response['error']['code'] == -32602


@pytest.mark.xfail
@pytest.mark.asyncio
async def test_install_with_uninitialised_profile(
    request: pytest.FixtureRequest,
    ws: ClientWebSocketResponse,
):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'install',
        'params': {
            'profile': request.node.name,
            'defns': [{'source': 'curse', 'name': 'molinari'}],
            'replace': False,
        },
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['error'] and rpc_response['error']['code'] == -32001



from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime
import typing
from typing import Any

from pydantic import BaseModel

from . import manager
from .config import Flavour
from .utils import bucketise, normalise_names


class _SameAs(BaseModel):
    source: str
    id: str


class BaseCatatalogueEntry(BaseModel):
    source: str
    id: str
    slug: str = ''
    name: str
    url: str
    game_flavours: typing.Set[Flavour]
    download_count: int
    last_updated: datetime
    folders: typing.List[typing.Set[str]] = []
    same_as: typing.List[_SameAs] = []


class CatalogueEntry(BaseCatatalogueEntry):
    normalised_name: str
    derived_download_score: float


class BaseCatalogue(BaseModel, json_encoders={set: sorted}):
    version = 5
    entries: typing.List[BaseCatatalogueEntry]

    @classmethod
    async def collate(cls, start_date: datetime | None):
        async with manager.init_web_client() as web_client:
            entries = [e for r in manager.Manager.RESOLVERS async for e in r.catalogue(web_client)]
            if start_date is not None:
                entries = [e for e in entries if e.last_updated >= start_date]
            return cls(entries=entries)


class Catalogue(BaseModel):
    version = 1
    entries: typing.List[CatalogueEntry]
    curse_slugs: typing.Dict[str, str]

    @classmethod
    def from_base_catalogue(cls, raw_base_catalogue: Sequence[Any], start_date: datetime | None):
        normalise = normalise_names('')

        base_entries = BaseCatalogue.parse_obj(raw_base_catalogue).entries
        if start_date is not None:
            base_entries = [e for e in base_entries if e.last_updated >= start_date]

        most_downloads_per_source = {
            s: max(e.download_count for e in i)
            for s, i in bucketise(base_entries, key=lambda e: e.source).items()
        }
        entries = [
            CatalogueEntry(
                **e.__dict__,
                normalised_name=normalise(e.name),
                derived_download_score=0
                if e.source == 'github'
                else e.download_count / most_downloads_per_source[e.source],
            )
            for e in base_entries
        ]
        return cls(
            entries=entries,
            curse_slugs={e.slug: e.id for e in entries if e.source == 'curse'},
        )



from __future__ import annotations

import importlib.util
import sys


def __getattr__(name: str) -> object:
    """Defer importing own modules until attempting to access an attribute.

    Importing this in ``__init__`` will overwrite relative imports.
    """
    fullname = __package__ + '.' + name
    try:
        return sys.modules[fullname]
    except KeyError:
        spec = importlib.util.find_spec(fullname)
        if spec is None:
            # ``AttributeError`` is converted to an ``ImportError`` by the import machinery
            raise AttributeError
        sys.modules[fullname] = module = importlib.util.module_from_spec(spec)
        assert spec.loader
        lazy_loader = importlib.util.LazyLoader(spec.loader)
        lazy_loader.exec_module(module)
        return module



from __future__ import annotations

from functools import partial
import json
import os

import click
from loguru import logger
import toga
import toga.constants
import toga.style

from . import json_rpc_server


class InstawowApp(toga.App):
    def __init__(self, **kwargs: object) -> None:
        super().__init__(
            formal_name='instawow-gui',
            app_id='org.instawow.instawow_gui',
            app_name='instawow_gui',
            icon='resources/instawow_gui',
            **kwargs,
        )

    def startup(self) -> None:
        self.main_window = toga.MainWindow(title=self.formal_name, size=(800, 600))
        if os.name == 'nt':
            import ctypes

            from . import webview2_adapter

            # Enable high DPI support.
            ctypes.windll.user32.SetProcessDPIAware()

            if not webview2_adapter.is_webview2_installed():
                self.main_window.content = toga.Box(
                    style=toga.style.Pack(
                        alignment=toga.constants.CENTER, direction=toga.constants.COLUMN
                    ),
                    children=[
                        toga.Label(
                            'WebView2 is required to run instawow.  Please install it '
                            'and restart instawow.',
                            style=toga.style.Pack(text_align=toga.constants.CENTER, padding=10),
                        ),
                        toga.Button(
                            'Download WebView2',
                            on_press=lambda _: click.launch(
                                'https://developer.microsoft.com/en-us/microsoft-edge/webview2/'
                            ),
                            style=toga.style.Pack(width=200),
                        ),
                    ],
                )
                self.main_window.show()
                return

            self.main_window.content = web_view = toga.WebView(
                style=toga.style.Pack(flex=1), factory=webview2_adapter.Factory
            )

        else:
            self.main_window.content = web_view = toga.WebView(style=toga.style.Pack(flex=1))

        def dispatch_js_keyboard_event(_: toga.Command, *, action: str):
            event_args = json.dumps(
                {'detail': {'action': action}, 'bubbles': True, 'cancelable': True}
            )
            web_view.invoke_javascript(
                f'document.dispatchEvent(new CustomEvent("togaSimulateKeypress", {event_args}));'
            )

        self.commands.add(
            toga.Command(
                partial(dispatch_js_keyboard_event, action='toggleSearchFilter'),
                label='Toggle Search Filter',
                shortcut=toga.Key.MOD_1 + toga.Key.G,
                group=toga.Group.EDIT,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewInstalled'),
                label='Installed',
                shortcut=toga.Key.MOD_1 + toga.Key.L,
                group=toga.Group.WINDOW,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewReconcile'),
                label='Unreconciled',
                group=toga.Group.WINDOW,
                section=20,
                order=20,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewSearch'),
                label='Search',
                shortcut=toga.Key.MOD_1 + toga.Key.F,
                group=toga.Group.WINDOW,
                section=20,
                order=30,
            ),
        )

        async def startup() -> None:
            web_app = await json_rpc_server.create_app(self.main_window)
            server_url, serve = await json_rpc_server.run_app(web_app)
            logger.debug(f'JSON-RPC server running on {server_url}')
            web_view.url = str(server_url)
            await serve()

        self._impl.loop.create_task(startup())

        self.main_window.show()



"""
Change the `pkg_options` table's `strategy` column to be
one of 'default', 'latest'.
"""
from functools import partial, update_wrapper

from alembic import op
from sqlalchemy import Enum, String
from sqlalchemy.sql import column, table

# revision identifiers, used by Alembic.
revision = 'e4ae835a34be'
down_revision = None
branch_labels = None
depends_on = None


def _migrate(new_type: Enum, old_val: str, new_val: str) -> None:
    intermediate_type = String
    pkg_options = table('pkg_options', column('strategy', intermediate_type))

    with op.batch_alter_table('pkg_options') as batch_op:
        batch_op.alter_column('strategy', type_=intermediate_type)

    op.execute(
        pkg_options.update()
        .where(pkg_options.c.strategy == op.inline_literal(old_val))
        .values({'strategy': op.inline_literal(new_val)})
    )

    with op.batch_alter_table('pkg_options') as batch_op:
        batch_op.alter_column('strategy', type_=new_type)


upgrade = update_wrapper(
    partial(_migrate, Enum('default', 'latest'), 'canonical', 'default'), _migrate
)
downgrade = update_wrapper(
    partial(_migrate, Enum('canonical', 'latest'), 'default', 'canonical'), _migrate
)



"""
Add the package changelog column.

Revision ID: 764fa963cc71
Revises: e4921edb1154
Create Date: 2021-03-20 02:46:24.528320

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '764fa963cc71'
down_revision = 'e4921edb1154'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg', schema=None) as batch_op:
        batch_op.add_column(sa.Column('changelog_url', sa.String(), nullable=True))

    conn = op.get_bind()
    conn.execute('update pkg set changelog_url = "data:,"')

    with op.batch_alter_table('pkg', schema=None) as batch_op:
        batch_op.alter_column('changelog_url', nullable=False)


def downgrade():
    with op.batch_alter_table('pkg', schema=None) as batch_op:
        batch_op.drop_column('changelog_url')



from __future__ import annotations

from collections.abc import Iterable
from contextlib import suppress
from functools import total_ordering
import re

import sqlalchemy as sa
from typing_extensions import TypeAlias

from . import manager
from .common import Flavour
from .db import pkg_folder
from .resolvers import (
    CurseResolver,
    Defn,
    GithubResolver,
    InstawowResolver,
    TukuiResolver,
    WowiResolver,
)
from .utils import TocReader, bucketise, cached_property, merge_intersecting_sets, uniq

FolderAndDefnPairs: TypeAlias = 'list[tuple[list[AddonFolder], list[Defn]]]'


_source_toc_ids = {
    'X-Curse-Project-ID': CurseResolver.source,
    'X-Tukui-ProjectID': TukuiResolver.source,
    'X-WoWI-ID': WowiResolver.source,
}
_source_sort_order = [
    GithubResolver.source,
    CurseResolver.source,
    WowiResolver.source,
    TukuiResolver.source,
    InstawowResolver.source,
]
# See https://github.com/Stanzilla/WoWUIBugs/issues/68#issuecomment-830351390
_flavour_toc_suffixes = {
    Flavour.retail: [
        '_Mainline.toc',
        '-Mainline.toc',
        '.toc',
    ],
    Flavour.vanilla_classic: [
        '_Vanilla.toc',
        '-Vanilla.toc',
        '_Classic.toc',
        '-Classic.toc',
        '.toc',
    ],
    Flavour.burning_crusade_classic: [
        '_TBC.toc',
        '-TBC.toc',
        '_BCC.toc',
        '-BCC.toc',
        '.toc',
    ],
}


@total_ordering
class AddonFolder:
    def __init__(self, name: str, toc_reader: TocReader) -> None:
        self.name = name
        self.toc_reader = toc_reader

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.name}>'

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (self.__class__, str)):
            return NotImplemented
        return self.name == other

    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            other = other.name
        elif not isinstance(other, str):
            return NotImplemented
        return self.name < other

    @cached_property
    def defns_from_toc(self) -> frozenset[Defn]:
        return frozenset(
            Defn(s, i)
            for s, i in ((s, self.toc_reader[k]) for k, s in _source_toc_ids.items())
            if i
        )

    @cached_property
    def version(self) -> str:
        return self.toc_reader['Version', 'X-Packaged-Version', 'X-Curse-Packaged-Version'] or ''


def get_unreconciled_folders(manager: manager.Manager) -> Iterable[AddonFolder]:
    pkg_folders = manager.database.execute(sa.select(pkg_folder.c.name)).scalars().all()
    unreconciled_folder_paths = (
        p
        for p in manager.config.addon_dir.iterdir()
        if p.name not in pkg_folders and p.is_dir() and not p.is_symlink()
    )
    suppress_not_found_error = suppress(FileNotFoundError)
    for path in unreconciled_folder_paths:
        for suffix in _flavour_toc_suffixes[manager.config.game_flavour]:
            with suppress_not_found_error:
                toc_reader = TocReader.from_addon_path(path, suffix)
                yield AddonFolder(path.name, toc_reader)
                break


def get_unreconciled_folder_set(manager: manager.Manager) -> frozenset[AddonFolder]:
    return frozenset(get_unreconciled_folders(manager))


async def match_toc_source_ids(
    manager: manager.Manager, leftovers: frozenset[AddonFolder]
) -> FolderAndDefnPairs:
    addons_with_toc_source_ids = [a for a in sorted(leftovers) if a.defns_from_toc]
    merged_defns = list(
        merge_intersecting_sets(a.defns_from_toc for a in addons_with_toc_source_ids)
    )
    folders_grouped_by_overlapping_defns = bucketise(
        addons_with_toc_source_ids, lambda a: next(d for d in merged_defns if a.defns_from_toc & d)
    )
    return [
        (f, sorted(b, key=lambda d: _source_sort_order.index(d.source)))
        for b, f in folders_grouped_by_overlapping_defns.items()
    ]


async def match_folder_name_subsets(
    manager: manager.Manager, leftovers: frozenset[AddonFolder]
) -> FolderAndDefnPairs:
    def sort_key(value: tuple[frozenset[AddonFolder], Defn]):
        folders, defn = value
        return (-len(folders), _source_sort_order.index(defn.source))

    catalogue = await manager.synchronise()
    matches = [
        (frozenset(e for e in leftovers if e.name in f), Defn(i.source, i.id))
        for i in catalogue.entries
        if manager.config.game_flavour in i.game_flavours
        for f in i.folders
        if f <= leftovers
    ]
    merged_folders = list(merge_intersecting_sets(f for f, _ in matches))
    matches_grouped_by_overlapping_folder_names = bucketise(
        matches, lambda v: next(f for f in merged_folders if v[0] & f)
    )
    return [
        (sorted(f), uniq(d for _, d in sorted(b, key=sort_key)))
        for f, b in matches_grouped_by_overlapping_folder_names.items()
    ]


async def match_addon_names_with_folder_names(
    manager: manager.Manager, leftovers: frozenset[AddonFolder]
) -> FolderAndDefnPairs:
    def normalise(value: str):
        return re.sub(r'[^0-9A-Za-z]', '', value.casefold())

    catalogue = await manager.synchronise()
    addon_names_to_catalogue_entries = bucketise(
        catalogue.entries, key=lambda i: normalise(i.name)
    )
    matches = (
        (a, addon_names_to_catalogue_entries.get(normalise(a.name))) for a in sorted(leftovers)
    )
    return [([a], uniq(Defn(i.source, i.id) for i in m)) for a, m in matches if m]



import sys

from instawow.cli import main

if __name__ == '__main__':
    if getattr(sys, 'frozen', False):
        import multiprocessing

        multiprocessing.freeze_support()

        prog_name = sys.executable
    else:
        prog_name = None

    main(prog_name=prog_name)



from __future__ import annotations

from collections.abc import Iterable, Set
import json
import os
from pathlib import Path, PurePath
from tempfile import gettempdir
import typing
from typing import Any

import click
from loguru import logger
from pydantic import BaseModel, BaseSettings, Field, PydanticValueError, SecretStr, validator
from pydantic.env_settings import SettingsSourceCallable

from .common import Flavour
from .utils import trash


class _PathNotWritableDirectoryError(PydanticValueError):
    code = 'path.not_writable_directory'
    msg_template = '"{path}" is not a writable directory'


def _expand_path(value: Path):
    return Path(os.path.abspath(os.path.expanduser(value)))


def _is_writable_dir(value: Path):
    return value.is_dir() and os.access(value, os.W_OK)


def _ensure_dirs(dirs: Iterable[Path]):
    for dir_ in dirs:
        dir_.mkdir(exist_ok=True, parents=True)


def _write_config(config: _BaseSettings, fields: Set[str], **kwargs: object):
    json_output = config.json(include=fields, indent=2, **kwargs)
    config.config_file.write_text(json_output, encoding='utf-8')


def _read_config(config: _BaseSettings, missing_ok: bool = False):
    try:
        return json.loads(config.config_file.read_bytes())
    except FileNotFoundError:
        if missing_ok:
            default_config: dict[str, Any] = {}
            return default_config
        raise


def _customise_sources(
    init_settings: SettingsSourceCallable,
    env_settings: SettingsSourceCallable,
    file_secret_settings: SettingsSourceCallable,
):
    # Prioritise env vars
    return (env_settings, init_settings)


class _BaseSettings(BaseSettings, env_prefix='INSTAWOW_', customise_sources=_customise_sources):
    @property
    def config_file(self) -> Path:
        raise NotImplementedError


class _AccessTokens(BaseModel):
    github: typing.Optional[SecretStr]
    wago: typing.Optional[SecretStr]


def _encode_reveal_secret_str(value: object):
    if isinstance(value, BaseModel):
        return value.dict()
    elif isinstance(value, SecretStr):
        return value.get_secret_value()
    raise TypeError('Unencodable value', value)


class GlobalConfig(_BaseSettings):
    config_dir: Path = Field(default_factory=lambda: Path(click.get_app_dir('instawow')))
    temp_dir: Path = Field(default_factory=lambda: Path(gettempdir(), 'instawow'))
    auto_update_check: bool = True
    access_tokens: _AccessTokens = _AccessTokens(github=None, wago=None)

    @validator('config_dir', 'temp_dir')
    def _expand_path(cls, value: Path) -> Path:
        return _expand_path(value)

    def list_profiles(self) -> list[str]:
        "Get the names of the profiles contained in ``config_dir``."
        profiles = [c.parent.name for c in self.config_dir.glob('profiles/*/config.json')]
        return profiles

    @classmethod
    def read(cls) -> GlobalConfig:
        dummy_config = cls()
        maybe_config_values = _read_config(dummy_config, missing_ok=True)
        return cls(**maybe_config_values) if maybe_config_values is not None else dummy_config

    def ensure_dirs(self) -> GlobalConfig:
        _ensure_dirs(
            [
                self.config_dir,
                self.temp_dir,
                self.cache_dir,
            ]
        )
        return self

    def write(self) -> GlobalConfig:
        """Write the configuration on disk.

        ``write`` should only be called when configuring instawow.
        This means that environment overrides should only be persisted
        if made during configuration.
        """
        self.ensure_dirs()
        _write_config(
            self, {'auto_update_check', 'access_tokens'}, encoder=_encode_reveal_secret_str
        )
        return self

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / 'cache'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'


class Config(_BaseSettings):
    global_config: GlobalConfig
    profile: str = Field(min_length=1, strip_whitespace=True)
    addon_dir: Path
    game_flavour: Flavour

    @validator('addon_dir')
    def _validate_path_is_writable_dir(cls, value: Path) -> Path:
        value = _expand_path(value)
        if not _is_writable_dir(value):
            raise _PathNotWritableDirectoryError(path=value)
        return value

    @staticmethod
    def infer_flavour(folder: os.PathLike[str] | str) -> Flavour:
        tail = tuple(map(str.casefold, PurePath(folder).parts[-3:]))
        if len(tail) != 3 or tail[1:] != ('interface', 'addons'):
            return Flavour.retail
        elif tail[0] in {'_classic_era_', '_classic_era_ptr_'}:
            return Flavour.vanilla_classic
        elif tail[0] in {'_classic_', '_classic_beta_', '_classic_ptr_'}:
            return Flavour.burning_crusade_classic
        else:
            return Flavour.retail

    @classmethod
    def read(cls, global_config: GlobalConfig, profile: str) -> Config:
        dummy_config = cls.construct(global_config=global_config, profile=profile)
        config = cls(global_config=global_config, **_read_config(dummy_config))
        return config

    def ensure_dirs(self) -> Config:
        _ensure_dirs(
            [
                self.profile_dir,
                self.logging_dir,
                self.plugin_dir,
            ]
        )
        return self

    def write(self) -> Config:
        self.ensure_dirs()
        _write_config(self, {'addon_dir', 'game_flavour', 'profile'})
        return self

    def delete(self) -> None:
        trash((self.profile_dir,), dest=self.global_config.temp_dir, missing_ok=True)

    @property
    def profile_dir(self) -> Path:
        return self.global_config.config_dir / 'profiles' / self.profile

    @property
    def logging_dir(self) -> Path:
        return self.profile_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.profile_dir / 'plugins'

    @property
    def config_file(self) -> Path:
        return self.profile_dir / 'config.json'

    @property
    def db_file(self) -> Path:
        return self.profile_dir / 'db.sqlite'


def _patch_loguru_enqueue():
    import queue
    import threading
    from types import SimpleNamespace

    import loguru._handler

    # On some systems when instantiating multiprocessing constructs Python
    # starts the multiprocessing resource monitor.  This is done in a subprocess
    # with ``sys.executable`` as the first argument.  In briefcase
    # this points to the same executable that starts the app - there
    # isn't a separate Python executable.  So with every new resource
    # monitor that spawns so does a new copy of the app, ad infinitum.
    # Even when not using briefcase spawning a subprocess slows down start-up,
    # not least because it imports a second copy of the ``site`` module.
    # We replace these multiprocessing constructs with their threading
    # equivalents in loguru since loguru itself does not spawn a subprocesses
    # but creates a separate thread for its "enqueued" logger and we don't
    # use multiprocessing in instawow.
    # This will definitely not come back to bite us.
    setattr(
        loguru._handler,
        'multiprocessing',
        SimpleNamespace(SimpleQueue=queue.Queue, Event=threading.Event, Lock=threading.Lock),
    )


def _intercept_logging_module_calls(log_level: str):  # pragma: no cover
    import logging

    class InterceptHandler(logging.Handler):
        logging_filename = getattr(logging, '__file__', None)

        def emit(self, record: logging.LogRecord) -> None:
            # Get the corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where the logged message originated
            frame = logging.currentframe()
            depth = 2
            while frame and frame.f_code.co_filename == self.logging_filename:
                frame = frame.f_back
                depth += 1

            logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

    logging.basicConfig(handlers=[InterceptHandler()], level=log_level)


def setup_logging(
    logging_dir: Path, log_level: str = 'INFO', log_to_stderr: bool = False
) -> int | None:
    if __debug__:
        _intercept_logging_module_calls(log_level)

    if not log_to_stderr:
        _patch_loguru_enqueue()

        (handler_id,) = logger.configure(
            handlers=[
                {
                    'sink': logging_dir / 'error.log',
                    'level': log_level,
                    'rotation': '5 MB',
                    'retention': 5,  # Number of log files to keep
                    'enqueue': True,
                },
            ]
        )
        return handler_id
